unit CPRPArser;

interface
uses
  SysUtils;
(*
  Author : Jens Borrisholt
  E Mail : Jens@Borrisholt.com
  Date   : 10. October 2004
  What   : This component check where a "CPR Number" is valid.
           The algorithem used are the official one published on :
           http://www.cpr.dk/publikationer/pnr-notat%20ny%20skrift.htm#Opbygning

  A "CPR Number" is a Danish Social Security number, and therefor the scope
  of this component is Danish developpers. Ohters are ofcause welcome ;o)

  Due to the fact of the scope of this component, all comments are written in
  Danish, since I suppose that if you need this component, you can also read the
  Danish language !?! Correct me if I'm wrong !

  Also, for the author's protection, I want to make certain that everyone
  understands that there is no warranty for this free software.
  If the software is modified by someone else and passed on, I want its
  recipients to know that what they have is not the original, so that any problems
  introduced by others will not reflect on the original authors' reputations.

  This phase was 99% copied form  GNU GENERAL PUBLIC LICENSE Version 2, June 1991

  You may use this component in all kinds of software, but please do not
  remove my name from the component.
*)

type
  TParseError = procedure (const ErrorText : String) of object;
  TCPRParser = class
  private
    FCPRNummer: String;
    FormatSettings : TFormatSettings;
    FOnParseError: TParseError;
    FValid: Boolean;
    procedure SetOnParseError(const Value: TParseError);
  protected
    function CheckCpr : boolean; virtual;
    procedure InternalWriteError(const ErrorText : String); overload; virtual;
    procedure InternalWriteError(const ErrorText : String; const Args: array of const); overload; virtual;
    procedure SetCPRNummer(const Value: String); virtual;
    function ValidateDate : boolean; virtual;
  public
    constructor Create; overload;
  published
    property CPRNummer : String read FCPRNummer write SetCPRNummer;
    property OnParseError : TParseError read FOnParseError write SetOnParseError;
    property Valid : Boolean read FValid;
  end;

implementation

uses
  Windows;

{.$Define English_Errors}
{$Define Danish_Errors}
(*
  Definer enten English_Errors eller Danish_Errors, for at f fejlbeskeder p
  enten hhv. Engelsk eller Dansk.

  Selvom dette komponent kun har sigte Danske udviklere, kan komponentet godt
  blive brugt i en engelsksproget applikation, og man derfor vil have brug for
  de engelske fejlbeskeder.
*)

{$IFNDEF English_Errors}
{$IFNDEF Danish_Errors}
  ERROR : Du skal angive sprog !!!
{$ENDIF}
{$ENDIF}

{ TCPerParser }

constructor TCPRParser.Create;
  function MAKELANGID(const p, s : Word) : Word;
  begin
    Result := (s shl 10) or p;
  end;
begin
  inherited;
  {$IFDEF English_Errors}
  GetLocaleFormatSettings(MAKELANGID(LANG_ENGLISH, GetThreadLocale), FormatSettings);
  {$ENDIF}
  {$IFDEF Danish_Errors}
  GetLocaleFormatSettings(MAKELANGID(LANG_DANISH, GetThreadLocale), FormatSettings);
  {$ENDIF}
  FValid := True;
end;

function TCPRParser.CheckCpr: boolean;
var
  i, Sum : Integer;
const
  CheckSum : array[1..10] of byte = (4, 3, 2, 7, 6, 5, 4, 3, 2, 1);
  //  Checksums tallene i rkkeflge.
begin
  if Length(CPRNummer) <> 10 then
  {$IFDEF English_Errors}
    InternalWriteError('The length of the CPR Number must be 10 or 11 characters');
  {$ENDIF}
  {$IFDEF Danish_Errors}
    InternalWriteError('Lngden af CPR Nummeret skal vre 10 eller 11 karekterer');
  {$ENDIF}

  Result := FValid;
  if not Result then
    exit;

  for i := 1 to 10 do
    Result := Result and (CPRNummer[i] in ['0'..'9']);

  if not Result then
  begin
  {$IFDEF English_Errors}
    InternalWriteError('The CPR Number may only contain counters and a dash');
  {$ENDIF}
  {$IFDEF Danish_Errors}
    InternalWriteError('CPR Nummeret m kun best af tal og en bindestreg');
  {$ENDIF}
    Exit;
  end;

  Result := ValidateDate;
  if not Result then
    exit;

  (*
    Kontrolberegning af et nummer med kontrolciffer:

    De 10 tal               0    7    0    7    6    1    -    4    2    8   5
    ganget med              x    x    x    x    x    x         x    x    x   x
    hver deres vrdi        4    3    2    7    6    5         4    3    2   1
    giver                   0 + 21 +  0 + 49 + 36 +  5      + 16 +  6 + 16 + 5

    sammenlagt giver disse tal 154.
    154 delt med 11 giver 14 og ingen rest.
    Nr 11 gr op i tallet i dette tilflde 154 og der derved ingen rest findes,
    er kontrolberegningen i orden.

    Ellers er kombinationen af de 10 cifre i nummeret forkert.

    Ref : http://www.cpr.dk/publikationer/pnr-notat%20ny%20skrift.htm#Opbygning
  *)

  Sum := 0;
  for i := 1 to 10 do
    Inc(Sum, StrToIntDef(CPRNummer[i], 0) * CheckSum[i]);
    (*
      Hvert tal i CPR Nummeret ganges op med deres respektive checksum ciffer
      og summeres sammen i Sum
    *)

  if Sum mod 11 <> 0 then
  {$IFDEF English_Errors}
    InternalWriteError('The checksum is incorrect !!!');
  {$ENDIF}
  {$IFDEF Danish_Errors}
    InternalWriteError('Checksummen er ikke korrekt !!!');
  {$ENDIF}
  (*
    Undersg om 11 gr op i tallet, og der derved ingen rest findes.
    Dermed er kontrolberegningen i orden.
  *)

  Result := FValid;
end;

procedure TCPRParser.SetCPRNummer(const Value: String);
begin
  FValid := true;
  FCPRNummer := Value;
  FCPRNummer := StringReplace(FCPRNummer, ' ', '', [rfReplaceAll]);

  if (Length(CPRNummer) = 11) and (CPRNummer[7] <> '-') then
  {$IFDEF English_Errors}
    InternalWriteError('The CPR Number must be of the form ddmmyy-xxxx');
  {$ENDIF}
  {$IFDEF Danish_Errors}
    InternalWriteError('CPR Nummeret skal vre af typen ddmm-xxxx');
  {$ENDIF}

  if not FValid then
    exit;

  FCPRNummer := StringReplace(FCPRNummer, '-', '', [rfReplaceAll]);
  (*
    CPR nummerer renses for mellemrum og bindestreg
  *)

  CheckCpr;
end;

function TCPRParser.ValidateDate: boolean;
var
  Day, Month, Year, Number, Tmp : Integer;
const
  DaysInMonths : array[Boolean] of array[1..12] of byte =
  (
    (31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31),
    (31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31)
  );
begin
  Day    := StrToIntDef(copy(CPRNummer, 1, 2), -1);
  Month  := StrToIntDef(copy(CPRNummer, 3, 2), -1);
  Year   := StrToIntDef(copy(CPRNummer, 5, 2), -1);
  Number := StrToIntDef(copy(CPRNummer, 7, 4), -1);

  if not (Day in [1..31]) then
  {$IFDEF English_Errors}
    InternalWriteError('Day must be within the range of 1 to 31') ;
  {$ENDIF}
  {$IFDEF Danish_Errors}
    InternalWriteError('Dag skal vre mellem 1 og 31') ;
  {$ENDIF}

   if not (Month in [1..12]) then
  {$IFDEF English_Errors}
    InternalWriteError('Month must be within the range of 1 and 12') ;
  {$ENDIF}
  {$IFDEF Danish_Errors}
    InternalWriteError('Mned skal vre mellem 1 og 12') ;
  {$ENDIF}

       if (Year >= 00) and (Year <= 99) and (Number >= 0001) and (Number <= 0999) then
    Inc(Year, 1900)
  else if (Year >= 00) and (Year <= 99) and (Number >= 1000) and (Number <= 1999) then
    Inc(Year, 1900)
  else if (Year >= 00) and (Year <= 99) and (Number >= 2000) and (Number <= 2999) then
    Inc(Year, 1900)
  else if (Year >= 00) and (Year <= 99) and (Number >= 3000) and (Number <= 3999) then
    Inc(Year, 1900)
  else if (Year >= 00) and (Year <= 36) and (Number >= 4000) and (Number <= 4999) then
    Inc(Year, 2000)
  else if (Year >= 37) and (Year <= 99) and (Number >= 4000) and (Number <= 4999) then
    Inc(Year, 1900)
  else if (Year >= 00) and (Year <= 57) and (Number >= 5000) and (Number <= 5999) then
    Inc(Year, 2000)
  else if (Year >= 58) and (Year <= 99) and (Number >= 5000) and (Number <= 5999) then
    Inc(Year, 1800)
  else if (Year >= 00) and (Year <= 57) and (Number >= 6000) and (Number <= 6999) then
    Inc(Year, 2000)
  else if (Year >= 58) and (Year <= 99) and (Number >= 6000) and (Number <= 6999) then
    Inc(Year, 1800)
  else if (Year >= 00) and (Year <= 57) and (Number >= 7000) and (Number <= 7999) then
    Inc(Year, 2000)
  else if (Year >= 58) and (Year <= 99) and (Number >= 7000) and (Number <= 7999) then
    Inc(Year, 1800)
  else if (Year >= 00) and (Year <= 57) and (Number >= 8000) and (Number <= 8999) then
    Inc(Year, 2000)
  else if (Year >= 58) and (Year <= 99) and (Number >= 8000) and (Number <= 8999) then
    Inc(Year, 1800)
  else if (Year >= 00) and (Year <= 36) and (Number >= 9000) and (Number <= 9999) then
    Inc(Year, 2000)
  else if (Year >= 37) and (Year <= 99) and (Number >= 9000) and (Number <= 9999) then
    Inc(Year, 2000);
  (*
     Tabel for om rstallet tilhrrer 18-, 19- eller 2000 tallet
     Ref : http://www.cpr.dk/publikationer/pnr-notat%20ny%20skrift.htm#Opbygning
  *)

  if (Year < 1858) or (Year > 2057) then
  {$IFDEF English_Errors}
    InternalWriteError('Year must be winthin the range of 1858 and 2057');
  {$ENDIF}
  {$IFDEF Danish_Errors}
    InternalWriteError('r skal vre mellem 1858 og 2057');
  {$ENDIF}
  (*
    I flge standarten for CPR Nummere skal ret vre mellen [1858 .. 2057]
  *)

  Tmp := DaysInMonths[IsLeapYear(Year)][Month];
  if Day > Tmp then
  {$IFDEF English_Errors}
    InternalWriteError('There are only %d days in %s %d', [Tmp, FormatSettings.LongMonthNames[Month], Year]);
  {$ENDIF}
  {$IFDEF Danish_Errors}
    InternalWriteError('%s %d har kun %d dage', [FormatSettings.LongMonthNames[Month], Year, Tmp]);
  {$ENDIF}
  (*
    Nr ret er endeligt beregnet, kan det ogs bestemmes om det er skudr eller ej,

    Check for om Day er strer end det antal dage der er i den pgldende mned,
    det pgldende r. Eks. 29/2 1900 eller 31/4
  *)

  Result := FValid;
end;

procedure TCPRParser.InternalWriteError(const ErrorText: String);
begin
  FValid := false;
  if Assigned(FOnParseError) then
    FOnParseError(ErrorText);
end;

procedure TCPRParser.InternalWriteError(const ErrorText: String; const Args: array of const);
begin
  InternalWriteError(Format(ErrorText, Args));
end;

procedure TCPRParser.SetOnParseError(const Value: TParseError);
begin
  FOnParseError := Value;
end;

end.
